#include "cdf.h"
#include "oal.h"
#include "utils.h"

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

static int pgm;
static int overwrite;
static int cdaweb;
static char *output;
static char *template;

static char *epoch_type;
static char *epoch_column;
static double epoch_base_time;

static char *
pds_get_key_value(OBJDESC *desc, const char *key)
{
  KEYWORD *ks;
  for (ks = desc->first_keyword; ks != NULL; ks = ks->right_sibling)
    if (streq(ks->name, key))
      return ks->value;
  return NULL;
}

static char *
pds_dup_key_value(OBJDESC *desc, const char *key)
{
  char *val = pds_get_key_value(desc, key);
  return val == NULL ? NULL : strdup(val);
}

static long
pds_get_key_long(OBJDESC *desc, const char *key)
{
  char *end, *str = pds_get_key_value(desc, key);
  if (str == NULL)
    return 0;
  long value = strtol(str, &end, 10);
  return str == end ? 0 : value;
}

static double
pds_get_key_double(OBJDESC *desc, const char *key)
{
  char *end, *str = pds_get_key_value(desc, key);
  if (str == NULL)
    return 0;
  double value = strtod(str, &end);
  return str == end ? 0 : value;
}

static char *
unquote(char *s)
{
  if (s != NULL)
    {
      int len = strlen(s);
      if (s[0] == '"' && s[len - 1] == '"')
	{
	  memmove(s, s + 1, len - 2);
	  s[len - 2] = '\0';
	}
    }
  return s;
}

static int
ends_in(const char *string, const char *pattern)
{
  int str_len = strlen(string);
  int pat_len = strlen(pattern);

  if (str_len < pat_len)
    return 0;

  return !strcmp(pattern, string + str_len - pat_len);
}

static uint8_t
norm255(double min, double max, double px)
{
  int ix = 255 * ((px - min) / (max - min));
  return ix <= 0 ? 0 : (ix <= 255 ? ix : 255);
}

static void
path_replace_extension(char *path, const char *old_ext, const char *new_ext)
{
  char *last_dot = strrchr(path, '.');
  if (last_dot != NULL && strcasecmp(last_dot, old_ext) == 0)
    strcpy(last_dot, new_ext);
  else
    strcpy(path + strlen(path), new_ext);
}

static void
copy_file(const char *src_fn, const char *dst_fn, int overwrite)
{
  char buf[4096];
  int src_fd, dst_fd, len;
  int flags = O_CREAT | O_WRONLY | O_TRUNC | (overwrite ? 0 : O_EXCL);

  if ((src_fd = open(src_fn, O_RDONLY, 0)) < 0)
    fail(errno, "Can't open source file %s", src_fn);
  if ((dst_fd = open(dst_fn, flags, 0666)) < 0)
    fail(errno, "Can't open destination file %s", dst_fn);

  while ((len = read(src_fd, buf, sizeof(buf))) > 0)
    write(dst_fd, buf, len);

  close(dst_fd);
  close(src_fd);
}

static CDFid
cdf_create(const char *cdf_fn, int overwrite)
{
  CDFid cdf_id;

  if (template)
    {
      copy_file(template, cdf_fn, overwrite);
      cdf_fail_unless(CDFopenCDF(cdf_fn, &cdf_id),
	"can't open CDF file %s", cdf_fn);
      return cdf_id;
    }

  if (overwrite)
    {
      remove(cdf_fn); // In case of a corrupt CDF file.
      CDFstatus status = CDFopenCDF(cdf_fn, &cdf_id);
      if (status != NO_SUCH_CDF)
	cdf_fail_unless(status, "can't open CDF file %s", cdf_fn);
      if (status == CDF_OK)
	cdf_fail_unless(CDFdeleteCDF(cdf_id),
	  "can't delete CDF file %s", cdf_fn);
    }

  cdf_fail_unless(CDFcreateCDF((char *)cdf_fn,  &cdf_id),
    "can't create CDF file %s", cdf_fn);
  return cdf_id;
}

static void
cdf_close(CDFid cdf_id)
{
  cdf_fail_unless(CDFcloseCDF(cdf_id),
    "can't close CDF output file");
}

static void
cdf_set_global_attr(CDFid cdf_id, const char *key, const char *val)
{
  long attr_id;
  cdf_fail_unless(CDFcreateAttr(cdf_id, key, (long)GLOBAL_SCOPE, &attr_id),
    "can't create attribute %s = %s", key, val);
  cdf_fail_unless(CDFputAttrgEntry(cdf_id, attr_id,
      (long)0, CDF_CHAR, strlen(val), (void *)val),
    "can't set attribute value %s = %s", key, val);
}

static long
cdf_add_column(CDFid cdf_id, const char *name, long type, long len)
{
  long col_id, zero = 0, vary = VARY;
  if ((col_id = CDFvarNum(cdf_id, (char *)name)) < 0)
    cdf_fail_unless(CDFcreatezVar(cdf_id, name, type,
	len,         // strlen for strings, 1 otherwise
	zero, &zero, // to designate a scalar variable
	vary, &vary, // record and dimension variances
	&col_id),
      "can't create variable %s", name);
  return col_id;
}

static void
cdf_set_column_value(CDFid cdf_id, long col_id, long recno, void *val_addr)
{
  cdf_fail_unless(CDFputzVarData(cdf_id, col_id, recno, NULL, val_addr),
    "can't set value of record %ld in column %ld", recno, col_id);
}

static long
cdf_get_attr_id(CDFid cdf_id, const char *key)
{
  long attr_id = CDFgetAttrNum(cdf_id, (void *)key);
  if (attr_id < 0)
    cdf_fail_unless(CDFcreateAttr(cdf_id, key, VARIABLE_SCOPE, &attr_id),
      "can't create '%s' attribute", key);
  return attr_id;
}

static void
cdf_set_attr_val(CDFid cdf_id, long var_id, const char *key,
  long type, long len, void *addr)
{
  long attr_id = cdf_get_attr_id(cdf_id, (void *)key);
  cdf_fail_unless(CDFputAttrzEntry(cdf_id, attr_id, var_id, type, len, addr),
    "can't put attr %s, type=%ld, len=%ld, addr=%p", key, type, len, addr);
}

static void
cdf_set_var_attr(CDFid cdf_id, long var_id, const char *key, const char *val)
{
  long attr_id = cdf_get_attr_id(cdf_id, key);
  cdf_fail_unless(CDFputAttrzEntry(cdf_id, attr_id, var_id,
      CDF_CHAR, strlen(val), (void *)val),
    "can't put attr %s, val = %s", key, val);
}

static char *
cdf_get_attr_str(CDFid cdf_id, const char *name)
{
  long type, size, zero = 0;
  long attr = CDFgetAttrNum(cdf_id, (void *)name);
  if (attr < CDF_OK)
    return NULL;

  cdf_fail_unless(CDFinquireAttrgEntry(cdf_id, attr, zero, &type, &size),
    "Can't get attr attributes for %s", name);
  char *buf = malloc(size + 1);
  cdf_fail_unless(CDFgetAttrgEntry(cdf_id, attr, zero, buf),
    "Can't get attr %s", name);
  buf[size] = '\0';
  return buf;
}


static void
cdf_set_pds_attr(CDFid cdf_id, long var_id, const char *cdf_key,
  OBJDESC *desc, const char *pds_key)
{
  char *value = unquote(pds_dup_key_value(desc, pds_key));
  if (value != NULL)
    {
      cdf_set_var_attr(cdf_id, var_id, cdf_key, value);
      free(value);
    }
}

static void
cdf_add_pds_image(CDFid cdf_id, const char *cdf_fn, OBJDESC *image_desc)
{
  OA_OBJECT obj = OaReadObject(image_desc);
  if (obj == NULL)
    fail(0, "can't OaReadObject() on %s", image_desc->class_name);
  debug(2, "image = %p, size = %ld", obj, obj->size);

  long rows = pds_get_key_long(image_desc, "LINES");
  long cols = pds_get_key_long(image_desc, "LINE_SAMPLES");
  long bits = pds_get_key_long(image_desc, "SAMPLE_BITS");
  long bytes = bits <= 8 ? 1 : (bits <= 16 ? 2 : (bits <= 32 ? 4 : 8));
  long size = rows * cols * bytes;

  debug(2, "image: rows = %ld, cols = %ld, bits = %ld, size = %ld",
    rows, cols, bits, size);
  if (rows == 0 || cols == 0 || bits == 0)
    fail(0, "unspecified row (%ld), column (%ld), or depth (%ld)",
      rows, cols, bits);
  if (obj->size != size)
    fail(0, "image size discrepancy: %ld != %ld", obj->size, size);

  long *type = NULL;
  long itype[] = { 0, CDF_INT1, CDF_INT2, 0, CDF_INT4, 0, 0, 0, 0 };
  long utype[] = { 0, CDF_UINT1, CDF_UINT2, 0, CDF_UINT4, 0, 0, 0, 0 };
  long ftype[] = { 0, 0, 0, 0, CDF_REAL4, 0, 0, 0, CDF_REAL8 };
  char *sample_type = unquote(pds_dup_key_value(image_desc, "SAMPLE_TYPE"));
  if (sample_type == NULL)
    fail(0, "undefined image sample type");
  else if (streq(sample_type, "MSB_INTEGER"))
    type = itype;
  else if (streq(sample_type, "MSB_UNSIGNED_INTEGER"))
    type = utype;
  else if (streq(sample_type, "IEEE_REAL"))
    type = ftype;
  else
    fail(0, "unsupported image sample type: %s", sample_type);
  free(sample_type);

  long img_id, two = 2, dims[] = { rows, cols }, vary[] = { VARY, VARY };
  cdf_fail_unless(CDFcreatezVar(cdf_id, image_desc->class_name, type[bytes], 1L,
      two, dims, vary, &vary, &img_id),
    "can't create %ldx%ld image array", rows, cols);

  int r, c;
  for (r = 0; r < rows; r++)
    for (c = 0; c < cols; c++)
      {
	long rc[] = { r, c };
	void *addr = ((char *)obj->data_ptr) + bytes * (r * cols + c);
	cdf_fail_unless(CDFputzVarData(cdf_id, img_id, 0L, rc, addr),
	  "can't put image[%d,%d]", r, c);
      }

  cdf_set_var_attr(cdf_id, img_id, "VAR_TYPE", "data");
  cdf_set_var_attr(cdf_id, img_id, "FIELDNAM", image_desc->class_name);
  cdf_set_var_attr(cdf_id, img_id, "CATDESC", image_desc->class_name);
  cdf_set_var_attr(cdf_id, img_id, "DISPLAY_TYPE", "image");
  cdf_set_var_attr(cdf_id, img_id, "LABLAXIS", "col");
  // cdf_set_attr_val(cdf_id, img_id, "VALIDMIN", type[bytes], 1, min_p);
  // cdf_set_attr_val(cdf_id, img_id, "VALIDMAX", type[bytes], 1, max_p);
  // cdf_set_attr_val(cdf_id, img_id, "FILLVAL",  type[bytes], 1, fill_p);

  if (pgm)
    {
      int cdf_len = strlen(cdf_fn);
      char pgm_fn[cdf_len + strlen(image_desc->class_name) + 7];
      sprintf(pgm_fn, "%.*s-%s.pgm",
	cdf_len - (ends_in(cdf_fn, ".cdf") ? 4 : 0),
        cdf_fn, image_desc->class_name);

      double min = pds_get_key_double(image_desc, "MINIMUM");
      double max = pds_get_key_double(image_desc, "MAXIMUM");
      debug(2, "hmin=%g, hmax=%g", min, max);
      FILE *pgm_fp = fopen(pgm_fn, "w");
      if (pgm_fp == NULL)
	fail(errno, "can't create file %s", pgm_fn);
      fprintf(pgm_fp, "P5 %ld %ld 255\n", cols, rows);

      char *cp = (char *)obj->data_ptr;
      double amin = *((float *)cp);
      double amax = *((float *)cp);
      for (r = 0; r < rows; r++)
	for (c = 0; c < cols; c++)
	  {
	    double px = *((float *)cp);
	    uint8_t ix = norm255(min, max, px);
 	    fwrite(&ix, 1, 1, pgm_fp);
	    cp += bytes;

	    if (px < amin) amin = px;
	    if (amax < px) amax = px;
	  }
      fclose(pgm_fp);
      debug(2, "amin=%g, amax=%g", amin, amax);      
    }
}

// space_tidy -- returns a newly allocated duplicate string with
// matched leading and trailing double-quotes removed, leading and
// trailing whitespace removed, and interior runs of whitespace
// compressed to a single space character.
static char *
string_tidy(const char *u_str)
{
  int u_len = strlen(u_str);
  char *str = malloc(u_len + 1);
  const char *src = u_str;
  char *dst = str;

  if (u_str[0] == '"' && u_str[u_len - 1] == '"')
    src++;
  while (isspace(*src))
    src++;

  while (*src)
    {
      if (! isspace(*src))
	*dst++ = *src++;
      else
	{
	  *dst++ = ' ';
	  while (isspace(*src))
	    src++;
	}
    }

  if (dst > str && u_str[0] == '"' && dst[-1] == '"')
    *--dst = '\0';

  if (dst > str && dst[-1] == ' ')
    dst--;
  *dst = '\0';

  return str;
}

static char *
string_shorten(const char *str, int min, int max)
{
  int len = strlen(str);
  if (len < max)
    return strdup(str);

  for (len = max; min < len && str[len] != ',' && str[len] != ' '; len--)
    ;

  char *trunc = malloc(max + 10);
  sprintf(trunc, "%.*s...", len, str);
  return trunc;
}

static void
cdf_add_double_range(CDFid cdf_id, long col_id, long cdf_type,
  double min, double max, double fill)
{
  cdf_set_attr_val(cdf_id, col_id, "VALIDMIN", cdf_type, 1, &min);
  cdf_set_attr_val(cdf_id, col_id, "VALIDMAX", cdf_type, 1, &max);
  cdf_set_attr_val(cdf_id, col_id, "FILLVAL",  cdf_type, 1, &fill);
}

static void
cdf_add_iso_epochs(CDFid cdf_id, long col_id,
  int width, const char *data, int rows)
{
  double min_val, max_val, std_fill = -1.0e31;
  char buf[30] = "0000-01-01T00:00:00.000";
  
  int recno;
  for (recno = 0; recno < rows; recno++)
    {
      memcpy(buf, data, width);
      double val = parseEPOCH3(buf);
      cdf_set_column_value(cdf_id, col_id, recno, &val);
      if (recno == 0 || val < min_val) min_val = val;
      if (recno == 0 || max_val < val) max_val = val;
      data += width;
    }
  debug(3, "%s: %g .. %g", "Epoch", min_val, max_val);
  cdf_add_double_range(cdf_id, col_id, CDF_EPOCH, min_val, max_val, std_fill);
}

static double
seconds_to_cdf_epoch(double base_period, double offset)
{
  double whole_secs, millisecs;
  double unix_secs = base_period + offset; // compute secs since unix epoch
  modf(modf(unix_secs, &whole_secs) * 1000, &millisecs);

  struct tm bdt;
  time_t time = lrint(whole_secs);
  gmtime_r(&time, &bdt);

  return computeEPOCH(
    1900L + bdt.tm_year, 1L + bdt.tm_mon, (long)bdt.tm_mday,
    (long)bdt.tm_hour, (long)bdt.tm_min, (long)bdt.tm_sec, lrint(millisecs));
}

static void
cdf_add_apollo_epochs(CDFid cdf_id, long col_id,
  double base_period, double *data, int rows)
{
  int recno;
  double min_val, max_val, std_fill = -1.0e31;  
  for (recno = 0; recno < rows; recno++)
    {
      double val = seconds_to_cdf_epoch(data[recno], base_period);
      cdf_set_column_value(cdf_id, col_id, recno, &val);
      if (recno == 0 || val < min_val) min_val = val;
      if (recno == 0 || max_val < val) max_val = val;
    }
  debug(3, "%s: %g .. %g", "Epoch", min_val, max_val);
  cdf_add_double_range(cdf_id, col_id, CDF_EPOCH, min_val, max_val, std_fill);
}

static double
jd_to_cdf_epoch(double jd)
{
  return seconds_to_cdf_epoch(0, (jd - 2440587.5) * 86400);
}

static void
cdf_add_jd_epochs(CDFid cdf_id, long col_id, double *data, int rows)
{
  int recno;
  double min_val, max_val, std_fill = -1.0e31;  
  for (recno = 0; recno < rows; recno++)
    {
      double val = jd_to_cdf_epoch(data[recno]);
      cdf_set_column_value(cdf_id, col_id, recno, &val);
      if (recno == 0 || val < min_val) min_val = val;
      if (recno == 0 || max_val < val) max_val = val;
    }
  debug(3, "%s: %g .. %g", "Epoch", min_val, max_val);
  cdf_add_double_range(cdf_id, col_id, CDF_EPOCH, min_val, max_val, std_fill);
}

static void
cdf_add_doubles(CDFid cdf_id, long col_id, 
  const char *name, const char *fill_str, double *data, int rows)
{
  double min_val = 0, max_val = 0, fill_val = 0, std_fill = -1.0e31;
  if (fill_str != NULL)
    fill_val = atof(fill_str);

  int recno, have_min_max = FALSE;
  for (recno = 0; recno < rows; recno++)
    {
      double val = data[recno];
      if (fill_str != NULL && val == fill_val)
        { // FIX THIS: Comparing floats for equality.
          cdf_set_column_value(cdf_id, col_id, recno, &std_fill);
          continue;
        }
      cdf_set_column_value(cdf_id, col_id, recno, &val);
      if (! have_min_max) { min_val = max_val = val; have_min_max = TRUE; }
      if (val < min_val) min_val = val;
      if (max_val < val) max_val = val;
    }

  debug(3, "%s: %g .. %g (%s)", name, min_val, max_val, fill_str);
  if (have_min_max)
    {
      cdf_set_attr_val(cdf_id, col_id, "VALIDMIN", CDF_DOUBLE, 1, &min_val);
      cdf_set_attr_val(cdf_id, col_id, "VALIDMAX", CDF_DOUBLE, 1, &max_val);
    }
  cdf_set_attr_val(cdf_id, col_id, "FILLVAL", CDF_DOUBLE, 1, &std_fill);
}

static void
cdf_add_integers(CDFid cdf_id, long col_id, 
  const char *name, const char *fill_str, int *data, int rows)
{
  int min_val = 0, max_val = 0, fill_val = 0, std_fill = -2147483647 - 1;
  if (fill_str != NULL)
    fill_val = atoi(fill_str);

  int recno, have_min_max = FALSE;
  for (recno = 0; recno < rows; recno++)
    {
      int val = data[recno];
      if (fill_str != NULL && val == fill_val)
        {
          cdf_set_column_value(cdf_id, col_id, recno, &std_fill);
          continue;
        }
      cdf_set_column_value(cdf_id, col_id, recno, &val);
      if (! have_min_max) { min_val = max_val = val; have_min_max = TRUE; }
      if (val < min_val) min_val = val;
      if (max_val < val) max_val = val;
    }

  debug(3, "%s: %i .. %i (%s)", name, min_val, max_val, fill_str);
  if (have_min_max)
    {
      cdf_set_attr_val(cdf_id, col_id, "VALIDMIN", CDF_INT4, 1, &min_val);
      cdf_set_attr_val(cdf_id, col_id, "VALIDMAX", CDF_INT4, 1, &max_val);
    }
  cdf_set_attr_val(cdf_id, col_id, "FILLVAL", CDF_INT4, 1, &std_fill);
}

static void
cdf_add_pds_table(CDFid cdf_id, OBJDESC *table_desc, OBJDESC *root_desc)
{
  OA_OBJECT obj = OaReadObject(table_desc);
  if (obj == NULL)
    fail(0, "can't OaReadObject() on %s", table_desc->class_name);
  debug(2, "table = %p, size = %ld", obj, obj->size);

  long rows = pds_get_key_long(table_desc, "ROWS");
  long cols = pds_get_key_long(table_desc, "COLUMNS");
  long size = obj->size;
  debug(2, "table = %p: rows = %ld, cols = %ld, size = %ld",
    obj, rows, cols, size);

  // try to find an overall description of the dataset.  Look first in
  // the table element, then in the root element.
  char *text = pds_get_key_value(table_desc, "DESCRIPTION");
  if (text == NULL)
    text = pds_get_key_value(root_desc, "DESCRIPTION");
  if (text == NULL)
    warn(0, "No overall description available");
  else
    cdf_set_global_attr(cdf_id, "TEXT", text);

  // start filling in data for each column
  ODLTREE col_tree = obj->odltree->first_child;
  for (; col_tree != NULL; col_tree = col_tree->right_sibling)
    {
      long col_id;
      char *name = unquote(pds_dup_key_value(col_tree, "NAME"));
      char *format = unquote(pds_dup_key_value(col_tree, "FORMAT"));
      char *fill_str = pds_get_key_value(col_tree, "MISSING_CONSTANT");

      char *description = unquote(pds_dup_key_value(col_tree, "DESCRIPTION"));
      char *long_desc = string_tidy(description);
      char *short_desc = string_shorten(long_desc, 8, 20);

      OA_OBJECT col_data = OaGetSubTable(obj, 1, rows, &col_tree, 1);
      debug(2, "column %s, format %s", name, format);

      int column_is_epoch = FALSE;
      if (epoch_column != NULL && strcasecmp(epoch_column, name) == 0)
        {
          column_is_epoch = TRUE;
          col_id = cdf_add_column(cdf_id, "Epoch", CDF_EPOCH, 1);

          cdf_set_var_attr(cdf_id, col_id, "VAR_TYPE", "support_data");
          cdf_set_var_attr(cdf_id, col_id, "CATDESC", long_desc);
          cdf_set_var_attr(cdf_id, col_id, "FIELDNAM", short_desc);

          if (streq(epoch_type, "epoch"))
            cdf_add_iso_epochs(cdf_id, col_id, 
              atoi(format + 1), (char *)col_data->data_ptr, rows);
          else if (streq(epoch_type, "julian"))
            cdf_add_jd_epochs(cdf_id, col_id, 
              (double *)col_data->data_ptr, rows);
          else if (streq(epoch_type, "apollo"))
            cdf_add_apollo_epochs(cdf_id, col_id, epoch_base_time,
              (double *)col_data->data_ptr, rows);
        }

      if (*format == 'F')
        {
          col_id = cdf_add_column(cdf_id, name, CDF_DOUBLE, 1);
          cdf_add_doubles(cdf_id, col_id, name, fill_str,
            (double *)col_data->data_ptr, rows);
        }
      else if (*format == 'I')
        {
          col_id = cdf_add_column(cdf_id, name, CDF_INT4, 1);
          cdf_add_integers(cdf_id, col_id, name, fill_str,
            (int *)col_data->data_ptr, rows);
        }
      else
        {
          warn(0, "Skipping column %s due to format %s", name, format);
  	}

      cdf_set_var_attr(cdf_id, col_id, "VAR_NOTES", description);
      cdf_set_var_attr(cdf_id, col_id, "CATDESC", long_desc);
      cdf_set_var_attr(cdf_id, col_id, "FIELDNAM", short_desc);
      cdf_set_var_attr(cdf_id, col_id, "LABLAXIS", short_desc);
      cdf_set_var_attr(cdf_id, col_id, "FORMAT", format);
      cdf_set_pds_attr(cdf_id, col_id, "UNITS", col_tree, "UNIT");
      cdf_set_var_attr(cdf_id, col_id, "DEPEND_0", "Epoch");
      if (column_is_epoch)
          cdf_set_var_attr(cdf_id, col_id, "VAR_TYPE", "support_data");
      else
        {
          cdf_set_var_attr(cdf_id, col_id, "VAR_TYPE", "data");
          cdf_set_var_attr(cdf_id, col_id, "DISPLAY_TYPE", "time_series");
        }

      free(short_desc);
      free(long_desc);
      free(description);
      free(format);
      free(name);
    }
}

static char *
cdf_get_prefix(CDFid cdf_id, const char *name)
{
  char *str = cdf_get_attr_str(cdf_id, name);
  if (str == NULL)
    fail(0, "Can't get attr string for %s", name);
  char *gtr = strchr(str, '>');
  if (gtr == NULL)
    fail(0, "No greater-than sign in '%s'", str);
  *gtr = '\0';
  return str;
}

// cdf_add_cdaweb_globals -- derives and adds the CDAWeb-mandated
// global attributes Logical_source, Data_version, and
// Logical_file_id.  Generates and returns the CDAWeb-mandated
// filename for the CDF file.
static char *
cdf_add_cdaweb_globals(CDFid cdf_id)
{
  // With --cdaweb, these attributes must be supplied by a template file.
  char *name = cdf_get_prefix(cdf_id, "Source_name");
  char *type = cdf_get_prefix(cdf_id, "Data_type");
  char *desc = cdf_get_prefix(cdf_id, "Descriptor");
  char *prefix = malloc(strlen(name) + strlen(type) + strlen(desc) + 3);
  sprintf(prefix, "%s_%s_%s", name, type, desc);
  cdf_set_global_attr(cdf_id, "Logical_source", prefix);

  // Use the first 'Epoch' variable as the start time.
  double epoch;
  long zero = 0, yr, mon, day, hr, min, sec, msec;
  static int days_in[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  cdf_fail_unless(CDFgetzVarData(cdf_id, CDFgetVarNum(cdf_id, "Epoch"),
      zero, &zero, &epoch), "Can't get epoch value for filename");
  EPOCHbreakdown(epoch, &yr, &mon, &day, &hr, &min, &sec, &msec);

  // Round times past noon up to the next day.
  if (hr >= 12) day++;
  days_in[2] = (yr % 4) == 0 ? 29 : 28;
  if (day > days_in[mon]) { day = 1; mon++; }
  if (mon > 12) yr++;

  // Assume a version of "1".
  char *vnum = cdf_get_attr_str(cdf_id, "Data_version");
  if (vnum == NULL)
    {
      vnum = strdup("1");
      cdf_set_global_attr(cdf_id, "Data_version", vnum);
    }

  // Mash it all together to form the Logical_file_id and the filename.
  char *fn = malloc(strlen(prefix) + 20);
  sprintf(fn, "%s_%04ld%02ld%02ld_v%02d",
    prefix, yr, mon, day, vnum ? atoi(vnum) : 1);
  cdf_set_global_attr(cdf_id, "Logical_file_id", fn);
  strcat(fn, ".cdf");
  free(name); free(type); free(desc); free(prefix); free(vnum);
  return fn;
}

static void
pds_read_label(const char *fn, const char *cdf_fn)
{
  ODLTREE lab = OaParseLabelFile((char *)fn, NULL, ODL_EXPAND_STRUCTURE, 1);
  if (lab == NULL)
    fail(0, "can't parse label file %s", fn);
  lab = OaConvertLabel(lab);

  CDFid cdf_id = cdf_create(cdf_fn, overwrite);

  KEYWORD *ks;
  OBJDESC *root_desc, *desc;
  int obj_num = 1;
  while ((desc = OdlFindObjDesc(lab, "*", 0, 0, obj_num, ODL_TO_END)) != NULL)
    {
      int obj_class = OaGetObjectClass(desc);
      debug(4, "desc(%d): %p: %ld, %s(%d)",
        obj_num, desc, desc->level, desc->class_name, obj_class);

      for (ks = desc->first_keyword; ks != NULL; ks = ks->right_sibling)
        {
          debug(4, "    (%s) = (%s)", ks->name, ks->value);
          if (streq(ks->name, "DESCRIPTION"))
            {
              char *trunc = string_shorten(ks->value, 8, 20);
              debug(5, "    (%s) = (%s)", ks->name, trunc);
              free(trunc);
            }
        }

      if (streq("ROOT", desc->class_name)) // OA_UNKNOWN_CLASS
	{
	  root_desc = desc;
          for (ks = desc->first_keyword; ks != NULL; ks = ks->right_sibling)
            cdf_set_global_attr(cdf_id, ks->name, ks->value);
	}
      else if (obj_class == OA_TABLE)
      	{
          cdf_add_pds_table(cdf_id, desc, root_desc);
      	}
      else if (obj_class == OA_IMAGE)
      	{
          cdf_add_pds_image(cdf_id, cdf_fn, desc);
      	}
      else if (obj_class == OA_COLUMN)
        ; // silently ignored
      else
	warn(0, "Ignoring unexpected class: %s", desc->class_name);
      obj_num += 1;
    }

  if (! cdaweb)
    cdf_close(cdf_id);
  else
    {
      char *cdaw_fn = cdf_add_cdaweb_globals(cdf_id);
      cdf_close(cdf_id);
      rename(cdf_fn, cdaw_fn);
    }
}

static char *
cdf_make_path(const char *pds_path)
{
  char *cdf_path;
  debug(1, "pds_path = %s", pds_path);

  cdf_path = malloc((output ? strlen(output) + 1 : 0) + strlen(pds_path) + 5);

  if (output == NULL)
    path_replace_extension(strcpy(cdf_path, pds_path), ".lbl", ".cdf");
  else
    {
      struct stat sb;
      strcpy(cdf_path, output);
      if (stat(output, &sb) == 0 && S_ISDIR(sb.st_mode))
	{
	  char *last_slash = strrchr(pds_path, '/');
	  if (cdf_path[strlen(cdf_path) - 1] != '/')
	    strcat(cdf_path, "/");
	  strcat(cdf_path, last_slash ? last_slash + 1 : pds_path);
	  path_replace_extension(cdf_path, ".lbl", ".cdf");
	}	
    }

  debug(1, "cdf_path = %s", cdf_path);
  return cdf_path;
}

static char *option_chars = "HVDPOco:t:e:";
static struct option long_options[] =
  {
    { "help",          0, NULL, 'H' },
    { "version",       0, NULL, 'V' },
    { "debug",         0, NULL, 'D' },
    { "pgm",           0, NULL, 'P' },
    { "overwrite",     0, NULL, 'O' },
    { "cdaweb",        0, NULL, 'c' },
    { "output",        1, NULL, 'o' },
    { "template",      1, NULL, 't' },
    { "epoch",         1, NULL, 'e' },
    { NULL,            0, NULL,  0  }
  };

static void
help(void)
{
  printf(
    "%s: usage: %s [-%s] [files...]\n"
    "\n"
    "    -H, --help: describes options\n"
    "    -V, --version: shows version\n"
    "    -D, --debug: spew out reams of debug info\n"
    "    -P, --pgm: writes a portable grey map for each image file\n"
    "    -O, --overwrite: allows existing CDF files to be overwritten\n"
    "    -c, --cdaweb: derives the output filename and some globals\n"
    "    -o, --output=<fn>: specifies output directory or filename\n"
    "    -t, --template=<fn>: specifies a template CDF file\n"
    "    -e, --epoch=<fn>: specifies the column used for the 'Epoch'\n"
    , program_name, program_name, option_chars);
}

static void
process_options(int c)
{
  switch (c)
    {
    case 'V':
      printf("%s: version %s\n", program_name, VERSION);
      exit(EXIT_SUCCESS);
    case 'H':
      help();
      exit(EXIT_SUCCESS);
    case 'O':
      overwrite = 1;
      break;
    case 'D':
      debug_level++;
      break;
    case 'P':
      pgm = 1;
      break;
    case 'c':
      cdaweb = 1;
      break;
    case 'o':
      output = optarg;
      break;
    case 't':
      template = optarg;
      break;
    case 'e':
      epoch_type = optarg;
      if (streq("epoch", epoch_type))
	{
	  epoch_column = "EPOCH";
	}
      else if (streq("julian", epoch_type))
	{
	  epoch_column = "JULIAN_DATE";
	}
      else if (streq("apollo15", epoch_type))
	{
	  epoch_type = "apollo";
	  epoch_column = "GROUND_ELAPSED_TIME";
	  epoch_base_time = 49383240;
	}
      else if (streq("apollo16", epoch_type))
	{
	  epoch_type = "apollo";
	  epoch_column = "GROUND_ELAPSED_TIME";
	  epoch_base_time = 72294840;
	}
      else
	fail(0, "Bad epoch.  Must be epoch, julian, apollo15, or apollo16");
      break;
    default:
      exit(EXIT_FAILURE);
    }
}

int
main(int argc, char **argv)
{
  program_name = "pds-to-cdf";

  int c;
  while ((c = getopt_long(argc, argv, option_chars, long_options, NULL)) != -1)
    process_options(c);

  for (; optind < argc; optind++)
    {
      char *cdf_path = cdf_make_path(argv[optind]);
      pds_read_label(argv[optind], cdf_path);
      free(cdf_path);
    }

  return EXIT_SUCCESS;
}
